# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from osc_lib.command import command

from osc_placement import version


BASE_URL = '/traits'
RP_BASE_URL = '/resource_providers/{uuid}'
RP_TRAITS_URL = '/resource_providers/{uuid}/traits'
FIELDS = ('name',)


class ListTrait(command.Lister):

    """Return a list of valid trait strings.

    This command requires at least ``--os-placement-api-version 1.6``.
    """

    def get_parser(self, prog_name):
        parser = super(ListTrait, self).get_parser(prog_name)

        parser.add_argument(
            '--name',
            metavar='<name>',
            help=('A string to filter traits. The following options '
                  'are available: startswith operator filters the '
                  'traits whose name begins with a specific prefix, '
                  'e.g. name=startswith:CUSTOM, in operator filters '
                  'the traits whose name is in the specified list, '
                  'e.g. name=in:HW_CPU_X86_AVX,HW_CPU_X86_SSE, '
                  'HW_CPU_X86_INVALID_FEATURE.')
        )

        parser.add_argument(
            '--associated',
            action='store_true',
            help=('If this parameter is presented, the returned '
                  'traits will be those that are associated with at '
                  'least one resource provider.')
        )

        return parser

    @version.check(version.ge('1.6'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = BASE_URL
        params = {}
        if parsed_args.name:
            params['name'] = parsed_args.name
        if parsed_args.associated:
            params['associated'] = parsed_args.associated
        traits = http.request('GET', url, params=params).json()['traits']
        return FIELDS, [[t] for t in traits]


class ShowTrait(command.ShowOne):

    """Check if a trait name exists in this cloud.

    This command requires at least ``--os-placement-api-version 1.6``.
    """

    def get_parser(self, prog_name):
        parser = super(ShowTrait, self).get_parser(prog_name)

        parser.add_argument(
            'name',
            metavar='<name>',
            help='Name of the trait.'
        )

        return parser

    @version.check(version.ge('1.6'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = '/'.join([BASE_URL, parsed_args.name])

        http.request('GET', url)
        return FIELDS, [parsed_args.name]


class CreateTrait(command.Command):

    """Create a new custom trait.

    Custom traits must begin with the prefix ``CUSTOM_`` and contain only the
    letters A through Z, the numbers 0 through 9 and the underscore "_"
    character.

    This command requires at least ``--os-placement-api-version 1.6``.
    """

    def get_parser(self, prog_name):
        parser = super(CreateTrait, self).get_parser(prog_name)

        parser.add_argument(
            'name',
            metavar='<name>',
            help='Name of the trait.'
        )

        return parser

    @version.check(version.ge('1.6'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = '/'.join([BASE_URL, parsed_args.name])
        http.request('PUT', url)


class DeleteTrait(command.Command):

    """Delete the trait specified by {name}.

    This command requires at least ``--os-placement-api-version 1.6``.
    """

    def get_parser(self, prog_name):
        parser = super(DeleteTrait, self).get_parser(prog_name)

        parser.add_argument(
            'name',
            metavar='<name>',
            help='Name of the trait.'
        )

        return parser

    @version.check(version.ge('1.6'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = '/'.join([BASE_URL, parsed_args.name])

        http.request('DELETE', url)


class ListResourceProviderTrait(command.Lister):

    """List traits associated with the resource provider identified by {uuid}.

    This command requires at least ``--os-placement-api-version 1.6``.
    """

    def get_parser(self, prog_name):
        parser = super(ListResourceProviderTrait, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider.'
        )

        return parser

    @version.check(version.ge('1.6'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = RP_TRAITS_URL.format(uuid=parsed_args.uuid)
        traits = http.request('GET', url).json()['traits']
        return FIELDS, [[t] for t in traits]


class SetResourceProviderTrait(command.Lister):

    """Associate traits with the resource provider identified by {uuid}.

    All the associated traits will be replaced by the traits specified.

    This command requires at least ``--os-placement-api-version 1.6``.
    """

    def get_parser(self, prog_name):
        parser = super(SetResourceProviderTrait, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider.'
        )

        parser.add_argument(
            '--trait',
            metavar='<trait>',
            help='Name of the trait. May be repeated.',
            default=[],
            action='append'
        )

        return parser

    @version.check(version.ge('1.6'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = RP_BASE_URL.format(uuid=parsed_args.uuid)
        rp = http.request('GET', url).json()
        url = RP_TRAITS_URL.format(uuid=parsed_args.uuid)
        payload = {
            'resource_provider_generation': rp['generation'],
            'traits': parsed_args.trait
        }
        traits = http.request('PUT', url, json=payload).json()['traits']
        return FIELDS, [[t] for t in traits]


class DeleteResourceProviderTrait(command.Command):

    """Dissociate all the traits from the resource provider.

    Note that this command is not atomic if multiple processes are managing
    traits for the same provider.

    This command requires at least ``--os-placement-api-version 1.6``.
    """

    def get_parser(self, prog_name):
        parser = super(DeleteResourceProviderTrait, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider.'
        )

        return parser

    @version.check(version.ge('1.6'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = RP_TRAITS_URL.format(uuid=parsed_args.uuid)
        http.request('DELETE', url)



# coding: utf-8
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

import oslotest.base as base

import osc_placement.resources.common as common


class TestCommon(base.BaseTestCase):
    def test_encode(self):
        self.assertEqual(u'привет'.encode('utf-8'),
                         common.encode(u'привет'))

    def test_encode_custom_encoding(self):
        self.assertEqual(u'привет'.encode('utf-16'),
                         common.encode(u'привет', 'utf-16'))

    def test_encode_non_string(self):
        self.assertEqual(b'bytesvalue',
                         common.encode(b'bytesvalue'))

    def test_url_with_filters(self):
        base_url = '/resource_providers'
        expected = '/resource_providers?name=test&uuid=123456'

        filters = collections.OrderedDict([('name', 'test'), ('uuid', 123456)])

        actual = common.url_with_filters(base_url, filters)
        self.assertEqual(expected, actual)

    def test_url_with_filters_empty(self):
        base_url = '/resource_providers'

        self.assertEqual(base_url, common.url_with_filters(base_url))
        self.assertEqual(base_url, common.url_with_filters(base_url, {}))

    def test_url_with_filters_unicode_string(self):
        base_url = '/resource_providers'
        expected = ('/resource_providers?'
                    'name=%D0%BF%D1%80%D0%B8%D0%B2%D0%B5%D1%82')

        actual = common.url_with_filters(base_url, {'name': u'привет'})
        self.assertEqual(expected, actual)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import operator
import uuid

from osc_placement.tests.functional import base


class TestResourceProvider(base.BaseTestCase):
    def test_resource_provider_create(self):
        name = self.rand_name('test_rp_creation')
        created = self.resource_provider_create(name)
        self.assertEqual(name, created['name'])

        retrieved = self.resource_provider_show(created['uuid'])
        self.assertEqual(created, retrieved)

    def test_resource_provider_delete(self):
        created = self.resource_provider_create()

        before_delete = self.resource_provider_list(uuid=created['uuid'])
        self.assertEqual([created['uuid']],
                         [rp['uuid'] for rp in before_delete])

        self.resource_provider_delete(created['uuid'])
        after_delete = self.resource_provider_list(uuid=created['uuid'])
        self.assertEqual([], after_delete)

    def test_resource_provider_delete_not_found(self):
        rp_uuid = str(uuid.uuid4())
        msg = 'No resource provider with uuid ' + rp_uuid + ' found'

        exc = self.assertRaises(base.CommandException,
                                self.resource_provider_delete, rp_uuid)
        self.assertIn(msg, str(exc))

    def test_resource_provider_set(self):
        orig_name = self.rand_name('test_rp_orig_name')
        created = self.resource_provider_create(name=orig_name)

        before_update = self.resource_provider_show(created['uuid'])
        self.assertEqual(orig_name, before_update['name'])
        self.assertEqual(0, before_update['generation'])

        new_name = self.rand_name('test_rp_new_name')
        self.resource_provider_set(created['uuid'], name=new_name)
        after_update = self.resource_provider_show(created['uuid'])
        self.assertEqual(new_name, after_update['name'])
        self.assertEqual(0, after_update['generation'])

    def test_resource_provider_set_not_found(self):
        rp_uuid = str(uuid.uuid4())
        msg = 'No resource provider with uuid ' + rp_uuid + ' found'

        exc = self.assertRaises(base.CommandException,
                                self.resource_provider_set, rp_uuid, 'test')
        self.assertIn(msg, str(exc))

    def test_resource_provider_show(self):
        created = self.resource_provider_create()

        retrieved = self.resource_provider_show(created['uuid'])
        self.assertEqual(created, retrieved)

    def test_resource_provider_show_allocations(self):
        consumer_uuid = str(uuid.uuid4())
        allocs = {consumer_uuid: {'resources': {'VCPU': 2}}}

        created = self.resource_provider_create()
        self.resource_inventory_set(created['uuid'],
                                    'VCPU=4', 'VCPU:max_unit=4')
        self.resource_allocation_set(
            consumer_uuid,
            ['rp={},VCPU=2'.format(created['uuid'])])

        expected = dict(created, allocations=allocs, generation=2)
        retrieved = self.resource_provider_show(created['uuid'],
                                                allocations=True)
        self.assertEqual(expected, retrieved)

    def test_resource_provider_show_allocations_empty(self):
        created = self.resource_provider_create()

        expected = dict(created, allocations={})
        retrieved = self.resource_provider_show(created['uuid'],
                                                allocations=True)
        self.assertEqual(expected, retrieved)

    def test_resource_provider_show_not_found(self):
        rp_uuid = str(uuid.uuid4())
        msg = 'No resource provider with uuid ' + rp_uuid + ' found'

        exc = self.assertRaises(base.CommandException,
                                self.resource_provider_show, rp_uuid)
        self.assertIn(msg, str(exc))

    def test_resource_provider_list(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()

        expected_full = sorted([rp1, rp2], key=operator.itemgetter('uuid'))
        self.assertEqual(
            expected_full,
            sorted([rp for rp in self.resource_provider_list()
                    if rp['name'] in (rp1['name'], rp2['name'])],
                   key=operator.itemgetter('uuid'))
        )

    def test_resource_provider_list_by_name(self):
        rp1 = self.resource_provider_create()
        self.resource_provider_create()

        expected_filtered_by_name = [rp1]
        self.assertEqual(
            expected_filtered_by_name,
            [rp for rp in self.resource_provider_list(name=rp1['name'])]
        )

    def test_resource_provider_list_by_uuid(self):
        rp1 = self.resource_provider_create()
        self.resource_provider_create()

        expected_filtered_by_uuid = [rp1]
        self.assertEqual(
            expected_filtered_by_uuid,
            [rp for rp in self.resource_provider_list(uuid=rp1['uuid'])]
        )

    def test_resource_provider_list_empty(self):
        by_name = self.resource_provider_list(name='some_non_existing_name')
        self.assertEqual([], by_name)

        by_uuid = self.resource_provider_list(uuid=str(uuid.uuid4()))
        self.assertEqual([], by_uuid)

    def test_fail_if_incorrect_options(self):
        # aggregate_uuids param is available starting 1.3
        self.assertCommandFailed(
            'Operation or argument is not supported',
            self.resource_provider_list, aggregate_uuids=['1'])
        # resource param is available starting 1.4
        self.assertCommandFailed('Operation or argument is not supported',
                                 self.resource_provider_list, resources=['1'])


class TestResourceProvider14(base.BaseTestCase):
    VERSION = '1.4'

    def test_return_empty_list_for_nonexistent_aggregate(self):
        self.resource_provider_create()
        agg = str(uuid.uuid4())
        rps, warning = self.resource_provider_list(aggregate_uuids=[agg],
                                                   may_print_to_stderr=True)
        self.assertEqual([], rps)
        self.assertIn('The --aggregate-uuid option is deprecated, '
                      'please use --member-of instead.', warning)

        # validate --member_of option works as expected
        rps = self.resource_provider_list(member_of=[agg])
        self.assertEqual([], rps)

    def test_return_properly_for_aggregate_uuid_request(self):
        self.resource_provider_create()
        rp2 = self.resource_provider_create()
        agg = str(uuid.uuid4())
        self.resource_provider_aggregate_set(rp2['uuid'], agg)
        rps, warning = self.resource_provider_list(
            aggregate_uuids=[agg, str(uuid.uuid4())], may_print_to_stderr=True)
        self.assertEqual(1, len(rps))
        self.assertEqual(rp2['uuid'], rps[0]['uuid'])
        self.assertIn('The --aggregate-uuid option is deprecated, '
                      'please use --member-of instead.', warning)

        # validate --member_of option works as expected
        rps = self.resource_provider_list(member_of=[agg])
        self.assertEqual(1, len(rps))
        self.assertEqual(rp2['uuid'], rps[0]['uuid'])

    def test_return_empty_list_if_no_resource(self):
        rp = self.resource_provider_create()
        self.assertEqual([], self.resource_provider_list(
            resources=['MEMORY_MB=256'], uuid=rp['uuid']))

    def test_return_properly_for_resource_request(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()
        self.resource_inventory_set(rp1['uuid'], 'PCI_DEVICE=8')
        self.resource_inventory_set(rp2['uuid'], 'PCI_DEVICE=16')
        rps = self.resource_provider_list(resources=['PCI_DEVICE=16'])
        self.assertEqual(1, len(rps))
        self.assertEqual(rp2['uuid'], rps[0]['uuid'])


class TestResourceProvider114(base.BaseTestCase):
    VERSION = '1.14'

    def test_resource_provider_create(self):
        created = self.resource_provider_create()
        self.assertIn('root_provider_uuid', created)
        self.assertIn('parent_provider_uuid', created)

    def test_resource_provider_set(self):
        created = self.resource_provider_create()
        updated = self.resource_provider_set(
            created['uuid'], name='some_new_name')
        self.assertIn('root_provider_uuid', updated)
        self.assertIn('parent_provider_uuid', updated)

    def test_resource_provider_show(self):
        created = self.resource_provider_create()
        retrieved = self.resource_provider_show(created['uuid'])
        self.assertIn('root_provider_uuid', retrieved)
        self.assertIn('parent_provider_uuid', retrieved)

    def test_resource_provider_list(self):
        self.resource_provider_create()
        retrieved = self.resource_provider_list()[0]
        self.assertIn('root_provider_uuid', retrieved)
        self.assertIn('parent_provider_uuid', retrieved)

    def test_resource_provider_create_with_parent(self):
        parent = self.resource_provider_create()
        child = self.resource_provider_create(
            parent_provider_uuid=parent['uuid'])
        self.assertEqual(child['parent_provider_uuid'], parent['uuid'])

    def test_resource_provider_create_then_set_parent(self):
        parent = self.resource_provider_create()
        wannabe_child = self.resource_provider_create()
        child = self.resource_provider_set(
            wannabe_child['uuid'],
            name='mandatory_name_1',
            parent_provider_uuid=parent['uuid'])
        self.assertEqual(child['parent_provider_uuid'], parent['uuid'])

    def test_resource_provider_set_reparent(self):
        parent1 = self.resource_provider_create()
        parent2 = self.resource_provider_create()
        child = self.resource_provider_create(
            parent_provider_uuid=parent1['uuid'])
        exc = self.assertRaises(
            base.CommandException,
            self.resource_provider_set,
            child['uuid'],
            name='mandatory_name_2',
            parent_provider_uuid=parent2['uuid'])
        self.assertIn('HTTP 400', str(exc))

    def test_resource_provider_list_in_tree(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create(
            parent_provider_uuid=rp1['uuid'])
        rp3 = self.resource_provider_create(
            parent_provider_uuid=rp1['uuid'])
        self.resource_provider_create()  # not in-tree
        retrieved = self.resource_provider_list(in_tree=rp2['uuid'])
        self.assertEqual(
            set([rp['uuid'] for rp in retrieved]),
            set([rp1['uuid'], rp2['uuid'], rp3['uuid']])
        )

    def test_resource_provider_delete_parent(self):
        parent = self.resource_provider_create()
        self.resource_provider_create(parent_provider_uuid=parent['uuid'])
        exc = self.assertRaises(
            base.CommandException,
            self.resource_provider_delete,
            parent['uuid'])
        self.assertIn('HTTP 409', str(exc))


class TestResourceProvider118(base.BaseTestCase):
    VERSION = '1.18'

    # NOTE(cdent): The choice of traits here is important. We need to
    # make sure that we do not overlap with 'test_show_required_trait'
    # in TestAllocationCandidate117 which also creates some resource
    # providers. In a multi-process enviromment, the tests can race.
    def test_show_required_trait(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()
        self.resource_inventory_set(
            rp1['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        self.resource_inventory_set(
            rp2['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        self.resource_provider_trait_set(
            rp1['uuid'], 'STORAGE_DISK_SSD', 'HW_NIC_SRIOV_MULTIQUEUE')
        self.resource_provider_trait_set(
            rp2['uuid'], 'STORAGE_DISK_HDD', 'HW_NIC_SRIOV_MULTIQUEUE')

        rps = self.resource_provider_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=('HW_NIC_SRIOV_MULTIQUEUE',))

        uuids = [rp['uuid'] for rp in rps]

        self.assertEqual(2, len(rps))
        self.assertIn(rp1['uuid'], uuids)
        self.assertIn(rp2['uuid'], uuids)

        # Narrow the results and check multiple args.
        rps = self.resource_provider_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=('STORAGE_DISK_HDD', 'HW_NIC_SRIOV_MULTIQUEUE',))

        self.assertEqual(1, len(rps))
        self.assertEqual(rp2['uuid'], rps[0]['uuid'])

    # Specifying forbidden traits weren't available until version 1.22
    def test_fail_if_forbidden_trait_wrong_version(self):
        self.assertCommandFailed(
            'Operation or argument is not supported with version 1.18',
            self.resource_provider_list,
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            forbidden=('STORAGE_DISK_HDD',))


class TestResourceProvider122(base.BaseTestCase):
    VERSION = '1.22'

    def test_hide_forbidden_trait(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()
        rp3 = self.resource_provider_create()
        self.resource_inventory_set(
            rp1['uuid'], 'MEMORY_MB=1024', 'DISK_GB=256')
        self.resource_inventory_set(
            rp2['uuid'], 'MEMORY_MB=1024', 'DISK_GB=256')
        self.resource_inventory_set(
            rp3['uuid'], 'MEMORY_MB=1024', 'DISK_GB=256')
        self.resource_provider_trait_set(
            rp1['uuid'], 'STORAGE_DISK_SSD', 'HW_CPU_X86_VMX')
        self.resource_provider_trait_set(
            rp2['uuid'], 'STORAGE_DISK_HDD', 'HW_CPU_X86_VMX')
        self.resource_provider_trait_set(
            rp3['uuid'], 'STORAGE_DISK_HDD', 'HW_CPU_X86_VMX')

        rps = self.resource_provider_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=('HW_CPU_X86_VMX',),
            forbidden=('STORAGE_DISK_HDD',))

        self.assertEqual(1, len(rps))
        self.assertEqual(rp1['uuid'], rps[0]['uuid'])

        rps = self.resource_provider_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=('HW_CPU_X86_VMX',),
            forbidden=('STORAGE_DISK_SSD',))

        uuids = [rp['uuid'] for rp in rps]

        self.assertEqual(2, len(uuids))
        self.assertNotIn(rp1['uuid'], uuids)
        self.assertIn(rp2['uuid'], uuids)
        self.assertIn(rp3['uuid'], uuids)

    def test_list_required_trait_any_trait_old_microversion(self):
        self.assertCommandFailed(
            'Operation or argument is not supported with version 1.22',
            self.resource_provider_list,
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=(
                'STORAGE_DISK_HDD,STORAGE_DISK_SSD',
                'HW_NIC_SRIOV_MULTIQUEUE'),
        )


class TestResourceProvider139(base.BaseTestCase):
    VERSION = '1.39'

    def test_list_required_trait_any_trait(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()
        self.resource_inventory_set(
            rp1['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        self.resource_inventory_set(
            rp2['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        self.resource_provider_trait_set(
            rp1['uuid'], 'STORAGE_DISK_SSD', 'HW_NIC_SRIOV_MULTIQUEUE')
        self.resource_provider_trait_set(
            rp2['uuid'], 'STORAGE_DISK_HDD', 'HW_NIC_SRIOV_MULTIQUEUE')

        rps = self.resource_provider_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=('HW_NIC_SRIOV_MULTIQUEUE',))

        self.assertEqual(
            {rp1['uuid'], rp2['uuid']}, {rp['uuid'] for rp in rps})

        # Narrow the results and check multiple args.
        rps = self.resource_provider_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=('STORAGE_DISK_HDD', 'HW_NIC_SRIOV_MULTIQUEUE',))

        self.assertEqual({rp2['uuid']}, {rp['uuid'] for rp in rps})

        # Query for (HDD or SSD) and MULTIQUEUE and see that both RP returned
        # again
        rps = self.resource_provider_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=(
                'STORAGE_DISK_HDD,STORAGE_DISK_SSD',
                'HW_NIC_SRIOV_MULTIQUEUE')
        )

        self.assertEqual(
            {rp1['uuid'], rp2['uuid']}, {rp['uuid'] for rp in rps})

        # Query for (HDD or SSD) and MULTIQUEUE and !SSD and see that one of
        # the RPs are filtered
        rps = self.resource_provider_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=(
                'STORAGE_DISK_HDD,STORAGE_DISK_SSD',
                'HW_NIC_SRIOV_MULTIQUEUE'),
            forbidden=('STORAGE_DISK_SSD',),
        )

        self.assertEqual({rp2['uuid']}, {rp['uuid'] for rp in rps})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import keystoneauth1.exceptions.http as ks_exceptions
import osc_lib.exceptions as exceptions
import oslotest.base as base
import requests
import simplejson as json

from osc_placement import http
from osc_placement import version

from oslo_serialization import jsonutils


class FakeResponse(requests.Response):
    def __init__(self, status_code, content=None, headers=None):
        super(FakeResponse, self).__init__()
        self.status_code = status_code
        if content:
            self._content = content
        if headers:
            self.headers = headers


class TestSessionClient(base.BaseTestCase):
    def test_wrap_http_exceptions(self):
        def go():
            with http._wrap_http_exceptions():
                error = {
                    "errors": [
                        {"status": 404,
                         "detail": ("The resource could not be found.\n\n"
                                    "No resource provider with uuid 123 "
                                    "found for delete")}
                    ]
                }
                response = mock.Mock(content=json.dumps(error))
                raise ks_exceptions.NotFound(response=response)

        exc = self.assertRaises(exceptions.NotFound, go)
        self.assertEqual(404, exc.http_status)
        self.assertIn('No resource provider with uuid 123 found',
                      str(exc))

    def test_unexpected_response(self):
        def go():
            with http._wrap_http_exceptions():
                raise ks_exceptions.InternalServerError()

        exc = self.assertRaises(ks_exceptions.InternalServerError, go)
        self.assertEqual(500, exc.http_status)
        self.assertIn('Internal Server Error (HTTP 500)', str(exc))

    def test_session_client_version(self):
        session = mock.Mock()
        ks_filter = {'service_type': 'placement',
                     'region_name': 'mock_region',
                     'interface': 'mock_interface'}

        # 1. target to a specific version
        target_version = '1.23'
        client = http.SessionClient(
            session, ks_filter, api_version=target_version)
        self.assertEqual(client.api_version, target_version)

        # validate that the server side is not called
        session.request.assert_not_called()

        # 2. negotiation succeeds and have the client's highest version
        target_version = '1'
        session.request.return_value = FakeResponse(200)
        client = http.SessionClient(
            session, ks_filter, api_version=target_version)
        self.assertEqual(client.api_version, version.MAX_VERSION_NO_GAP)

        # validate that the server side is called
        expected_version = 'placement ' + version.MAX_VERSION_NO_GAP
        expected_headers = {'OpenStack-API-Version': expected_version,
                            'Accept': 'application/json'}
        session.request.assert_called_once_with(
            '/', 'GET', endpoint_filter=ks_filter,
            headers=expected_headers, raise_exc=False)
        session.reset_mock()

        # 3. negotiation fails and get the servers's highest version
        mock_server_version = '1.10'
        json_mock = {
            "errors": [{"status": 406,
                        "title": "Not Acceptable",
                        "min_version": "1.0",
                        "max_version": mock_server_version}]
        }
        session.request.return_value = FakeResponse(
            406, content=jsonutils.dump_as_bytes(json_mock))

        client = http.SessionClient(
            session, ks_filter, api_version=target_version)
        self.assertEqual(client.api_version, mock_server_version)

        # validate that the server side is called
        session.request.assert_called_once_with(
            '/', 'GET', endpoint_filter=ks_filter,
            headers=expected_headers, raise_exc=False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import itertools

from osc_lib.command import command
from osc_lib import exceptions
from osc_lib.i18n import _
from osc_lib import utils
from oslo_utils import excutils

from osc_placement.resources import common
from osc_placement import version


BASE_URL = '/resource_providers/{uuid}/inventories'
PER_CLASS_URL = BASE_URL + '/{resource_class}'
RP_BASE_URL = '/resource_providers'
USAGES_BASE_URL = '/resource_providers/{uuid}/usages'
INVENTORY_FIELDS = {
    'allocation_ratio': {
        'type': float,
        'required': False,
        'help': ('It is used in determining whether consumption '
                 'of the resource of the provider can exceed '
                 'physical constraints. For example, for a vCPU resource '
                 'with: allocation_ratio = 16.0, total = 8. '
                 'Overall capacity is equal to 128 vCPUs.')
    },
    'min_unit': {
        'type': int,
        'required': False,
        'help': ('A minimum amount any single allocation against '
                 'an inventory can have.')
    },
    'max_unit': {
        'type': int,
        'required': False,
        'help': ('A maximum amount any single allocation against '
                 'an inventory can have.')
    },
    'reserved': {
        'type': int,
        'required': False,
        'help': ('The amount of the resource a provider has reserved '
                 'for its own use.')
    },
    'step_size': {
        'type': int,
        'required': False,
        'help': ('A representation of the divisible amount of the resource '
                 'that may be requested. For example, step_size = 5 means '
                 'that only values divisible by 5 (5, 10, 15, etc.) '
                 'can be requested.')
    },
    'total': {
        'type': int,
        'required': True,
        'help': ('The actual amount of the resource that the provider '
                 'can accommodate.')
    }
}
FIELDS = tuple(INVENTORY_FIELDS.keys())
RC_HELP = ('<resource_class> is an entity that indicates standard or '
           'deployer-specific resources that can be provided by a resource '
           'provider. For example, VCPU, MEMORY_MB, DISK_GB.')


def parse_resource_argument(resource):
    parts = resource.split('=')
    if len(parts) != 2:
        raise ValueError(
            'Resource argument must have "name=value" format')
    name, value = parts
    parts = name.split(':')
    if len(parts) == 2:
        name, field = parts
    elif len(parts) == 1:
        name = parts[0]
        field = 'total'
    else:
        raise ValueError('Resource argument can contain only one colon')
    if not all([name, field, value]):
        raise ValueError('Name, field and value must be not empty')
    if field not in INVENTORY_FIELDS:
        raise ValueError('Unknown inventory field %s' % field)
    value = INVENTORY_FIELDS[field]['type'](value)
    return name, field, value


class SetInventory(command.Lister, version.CheckerMixin):

    """Replaces the set of inventory records for the resource provider.

    Note that by default this is a full replacement of the existing inventory.
    If you want to retain the existing inventory and add a new resource class
    inventory, you must specify all resource class inventory, old and new, or
    specify the ``--amend`` option.

    If a specific inventory field is not specified for a given resource class,
    it is assumed to be the total, i.e. ``--resource VCPU=16`` is equivalent to
    ``--resource VCPU:total=16``.

    Example::

        openstack resource provider inventory set <uuid> \
            --resource VCPU=16 \
            --resource MEMORY_MB=2048 \
            --resource MEMORY_MB:step_size=128
    """

    def get_parser(self, prog_name):
        parser = super(SetInventory, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider or UUID of the aggregate, '
                 'if --aggregate is specified'
        )
        fields_help = '\n'.join(
            '{} - {}'.format(f, INVENTORY_FIELDS[f]['help'].lower())
            for f in INVENTORY_FIELDS)
        parser.add_argument(
            '--resource',
            metavar='<resource_class>:<inventory_field>=<value>',
            help='String describing resource.\n' + RC_HELP + '\n'
                 '<inventory_field> (optional) can be:\n' + fields_help,
            default=[],
            action='append'
        )
        parser.add_argument(
            '--aggregate',
            action='store_true',
            help='If this option is specified, the inventories for all '
                 'resource providers that are members of the aggregate will '
                 'be set. This option requires at least '
                 '``--os-placement-api-version 1.3``'
        )
        parser.add_argument(
            '--amend',
            action='store_true',
            help='If this option is specified, the inventories will be '
                 'amended instead of being fully replaced'
        )
        parser.add_argument(
            '--dry-run',
            action='store_true',
            help='If this option is specified, the inventories that would be '
                 'set will be returned without actually setting any '
                 'inventories'
        )

        return parser

    def take_action(self, parsed_args):

        http = self.app.client_manager.placement

        if parsed_args.aggregate:
            self.check_version(version.ge('1.3'))
            filters = {'member_of': parsed_args.uuid}
            url = common.url_with_filters(RP_BASE_URL, filters)
            rps = http.request('GET', url).json()['resource_providers']
            if not rps:
                raise exceptions.CommandError(
                    'No resource providers found in aggregate with uuid %s.' %
                    parsed_args.uuid)
        else:
            url = RP_BASE_URL + '/' + parsed_args.uuid
            rps = [http.request('GET', url).json()]

        resources_list = []
        ret = 0
        for rp in rps:
            inventories = collections.defaultdict(dict)
            url = BASE_URL.format(uuid=rp['uuid'])
            if parsed_args.amend:
                # Get existing inventories
                # TODO(melwitt): Do something to handle the possibility of the
                # GET failing here (example: resource provider deleted from
                # underneath us).
                payload = http.request('GET', url).json()
                inventories.update(payload['inventories'])
                payload['inventories'] = inventories
            else:
                payload = {'inventories': inventories,
                           'resource_provider_generation': rp['generation']}

            # Apply resource values to inventories
            for r in parsed_args.resource:
                name, field, value = parse_resource_argument(r)
                inventories[name][field] = value

            try:
                if not parsed_args.dry_run:
                    resources = http.request('PUT', url, json=payload).json()
                else:
                    resources = payload
            except Exception as exp:
                with excutils.save_and_reraise_exception() as err_ctx:
                    if parsed_args.aggregate:
                        self.log.error(_('Failed to set inventory for '
                                         'resource provider %(rp)s: %(exp)s.'),
                                       {'rp': rp['uuid'], 'exp': exp})
                        err_ctx.reraise = False
                        ret += 1
                        continue
            resources_list.append((rp['uuid'], resources))

        if ret > 0:
            msg = _('Failed to set inventory for %(ret)s of %(total)s '
                    'resource providers.') % {'ret': ret, 'total': len(rps)}
            raise exceptions.CommandError(msg)

        def get_rows(fields, resources, rp_uuid=None):
            inventories = [
                dict(resource_class=k, **v)
                for k, v in resources['inventories'].items()
            ]
            prepend = (rp_uuid, ) if rp_uuid else ()
            # This is a generator expression
            rows = (prepend + utils.get_dict_properties(i, fields)
                    for i in inventories)
            return rows

        fields = ('resource_class', ) + FIELDS
        if parsed_args.aggregate:
            # If this is an aggregate batch, create output that will include
            # resource provider as the first field to differentiate the values
            rows = ()
            for rp_uuid, resources in resources_list:
                subrows = get_rows(fields, resources, rp_uuid=rp_uuid)
                rows = itertools.chain(rows, subrows)
            fields = ('resource_provider', ) + fields
            return fields, rows
        else:
            # If this was not an aggregate batch, show output for the one
            # resource provider (show payload of the first item in the list),
            # keeping the behavior prior to the addition of --aggregate option
            return fields, get_rows(fields, resources_list[0][1])


class SetClassInventory(command.ShowOne):
    """Replace the inventory record of the class for the resource provider.

    Example::

        openstack resource provider inventory class set <uuid> VCPU \
            --total 16 \
            --max_unit 4 \
            --reserved 1
    """

    def get_parser(self, prog_name):
        parser = super(SetClassInventory, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )
        parser.add_argument(
            'resource_class',
            metavar='<class>',
            help=RC_HELP
        )
        for name, props in INVENTORY_FIELDS.items():
            parser.add_argument(
                '--' + name,
                metavar='<{}>'.format(name),
                required=props['required'],
                type=props['type'],
                help=props['help'])

        return parser

    def take_action(self, parsed_args):

        http = self.app.client_manager.placement

        url = RP_BASE_URL + '/' + parsed_args.uuid
        rp = http.request('GET', url).json()

        payload = {'resource_provider_generation': rp['generation']}
        for field in FIELDS:
            value = getattr(parsed_args, field, None)
            if value is not None:
                payload[field] = value

        url = PER_CLASS_URL.format(uuid=parsed_args.uuid,
                                   resource_class=parsed_args.resource_class)
        resource = http.request('PUT', url, json=payload).json()
        return FIELDS, utils.get_dict_properties(resource, FIELDS)


class DeleteInventory(command.Command, version.CheckerMixin):
    """Delete the inventory.

    Depending on the resource class argument presence, delete all inventory for
    a given resource provider or for a resource provider/class pair.

    Delete all inventories for given resource provider requires at least
    ``--os-placement-api-version 1.5``.
    """

    def get_parser(self, prog_name):
        parser = super(DeleteInventory, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )
        parser.add_argument(
            '--resource-class',
            metavar='<resource_class>',
            required=self.compare_version(version.lt('1.5')),
            help=(RC_HELP + '\n'
                  'This argument can be omitted starting with '
                  '``--os-placement-api-version 1.5``. If it is omitted all '
                  'inventories of the specified resource provider '
                  'will be deleted.')
        )

        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement
        url = BASE_URL
        params = {'uuid': parsed_args.uuid}
        if parsed_args.resource_class is not None:
            url = PER_CLASS_URL
            params = {'uuid': parsed_args.uuid,
                      'resource_class': parsed_args.resource_class}

        http.request('DELETE', url.format(**params))


class ShowInventory(command.ShowOne):

    """Show the inventory for a given resource provider/class pair."""

    def get_parser(self, prog_name):
        parser = super(ShowInventory, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )
        parser.add_argument(
            'resource_class',
            metavar='<resource_class>',
            help=RC_HELP
        )

        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = PER_CLASS_URL.format(uuid=parsed_args.uuid,
                                   resource_class=parsed_args.resource_class)
        resource = http.request('GET', url).json()

        # TODO(stephenfin): We should just include this information in the
        # above API. Alternatively, we should add an API to retrieve usage for
        # a single resource class
        url = USAGES_BASE_URL.format(uuid=parsed_args.uuid)
        resources = http.request('GET', url).json()['usages']

        resource['used'] = resources[parsed_args.resource_class]

        fields = FIELDS + ('used', )

        return fields, utils.get_dict_properties(resource, fields)


class ListInventory(command.Lister):

    """List inventories for a given resource provider."""

    def get_parser(self, prog_name):
        parser = super(ListInventory, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )

        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = BASE_URL.format(uuid=parsed_args.uuid)
        resources = http.request('GET', url).json()

        inventories = [
            dict(resource_class=k, **v)
            for k, v in resources['inventories'].items()
        ]

        # TODO(stephenfin): We should just include this information in the
        # above API
        url = USAGES_BASE_URL.format(uuid=parsed_args.uuid)
        resources = http.request('GET', url).json()['usages']

        for inventory in inventories:
            inventory['used'] = resources[inventory['resource_class']]

        fields = ('resource_class', ) + FIELDS + ('used', )
        rows = (utils.get_dict_properties(i, fields) for i in inventories)
        return fields, rows



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from osc_placement.tests.functional import base


class TestAggregate(base.BaseTestCase):
    VERSION = '1.1'

    def test_fail_if_no_rp(self):
        self.assertCommandFailed(
            base.ARGUMENTS_MISSING,
            self.openstack,
            'resource provider aggregate list')

    def test_fail_if_rp_not_found(self):
        self.assertCommandFailed(
            'No resource provider',
            self.resource_provider_aggregate_list,
            'fake-uuid')

    def test_return_empty_list_if_no_aggregates(self):
        rp = self.resource_provider_create()
        self.assertEqual(
            [], self.resource_provider_aggregate_list(rp['uuid']))

    def test_success_set_aggregate(self):
        rp = self.resource_provider_create()
        aggs = {str(uuid.uuid4()) for _ in range(2)}
        rows = self.resource_provider_aggregate_set(
            rp['uuid'], *aggs)

        self.assertEqual(aggs, {r['uuid'] for r in rows})
        rows = self.resource_provider_aggregate_list(rp['uuid'])
        self.assertEqual(aggs, {r['uuid'] for r in rows})
        self.resource_provider_aggregate_set(rp['uuid'])
        rows = self.resource_provider_aggregate_list(rp['uuid'])
        self.assertEqual([], rows)

    def test_set_aggregate_fail_if_no_rp(self):
        self.assertCommandFailed(
            base.ARGUMENTS_MISSING,
            self.openstack,
            'resource provider aggregate set')

    def test_success_set_multiple_aggregates(self):
        # each rp is associated with two aggregates
        rps = [self.resource_provider_create() for _ in range(2)]
        aggs = {str(uuid.uuid4()) for _ in range(2)}
        for rp in rps:
            rows = self.resource_provider_aggregate_set(rp['uuid'], *aggs)
            self.assertEqual(aggs, {r['uuid'] for r in rows})
        # remove association for the first aggregate
        rows = self.resource_provider_aggregate_set(rps[0]['uuid'])
        self.assertEqual([], rows)
        # second rp should be in aggregates
        rows = self.resource_provider_aggregate_list(rps[1]['uuid'])
        self.assertEqual(aggs, {r['uuid'] for r in rows})
        # cleanup
        rows = self.resource_provider_aggregate_set(rps[1]['uuid'])
        self.assertEqual([], rows)

    def test_success_set_large_number_aggregates(self):
        rp = self.resource_provider_create()
        aggs = {str(uuid.uuid4()) for _ in range(100)}
        rows = self.resource_provider_aggregate_set(
            rp['uuid'], *aggs)
        self.assertEqual(aggs, {r['uuid'] for r in rows})
        rows = self.resource_provider_aggregate_set(rp['uuid'])
        self.assertEqual([], rows)

    def test_fail_if_incorrect_aggregate_uuid(self):
        rp = self.resource_provider_create()
        aggs = ['abc', 'efg']
        self.assertCommandFailed(
            "is not a 'uuid'",
            self.resource_provider_aggregate_set,
            rp['uuid'], *aggs)

    # In version 1.1 a generation is not allowed.
    def test_fail_generation_arg_version_handling(self):
        rp = self.resource_provider_create()
        agg = str(uuid.uuid4())
        self.assertCommandFailed(
            "Operation or argument is not supported with version 1.1",
            self.resource_provider_aggregate_set,
            rp['uuid'], agg, generation=5)


class TestAggregate119(TestAggregate):
    VERSION = '1.19'

    def test_success_set_aggregate(self):
        rp = self.resource_provider_create()
        aggs = {str(uuid.uuid4()) for _ in range(2)}
        rows = self.resource_provider_aggregate_set(
            rp['uuid'], *aggs, generation=rp['generation'])

        self.assertEqual(aggs, {r['uuid'] for r in rows})
        rows = self.resource_provider_aggregate_list(rp['uuid'])
        self.assertEqual(aggs, {r['uuid'] for r in rows})
        self.resource_provider_aggregate_set(
            rp['uuid'], *[], generation=rp['generation'] + 1)
        rows = self.resource_provider_aggregate_list(rp['uuid'])
        self.assertEqual([], rows)

    def test_success_set_multiple_aggregates(self):
        # each rp is associated with two aggregates
        rps = [self.resource_provider_create() for _ in range(2)]
        aggs = {str(uuid.uuid4()) for _ in range(2)}
        for rp in rps:
            rows = self.resource_provider_aggregate_set(
                rp['uuid'], *aggs, generation=rp['generation'])
            self.assertEqual(aggs, {r['uuid'] for r in rows})
        # remove association for the first aggregate
        rows = self.resource_provider_aggregate_set(
            rps[0]['uuid'], *[], generation=rp['generation'] + 1)
        self.assertEqual([], rows)
        # second rp should be in aggregates
        rows = self.resource_provider_aggregate_list(rps[1]['uuid'])
        self.assertEqual(aggs, {r['uuid'] for r in rows})
        # cleanup
        rows = self.resource_provider_aggregate_set(
            rps[1]['uuid'], *[], generation=rp['generation'] + 1)
        self.assertEqual([], rows)

    def test_success_set_large_number_aggregates(self):
        rp = self.resource_provider_create()
        aggs = {str(uuid.uuid4()) for _ in range(100)}
        rows = self.resource_provider_aggregate_set(
            rp['uuid'], *aggs, generation=rp['generation'])
        self.assertEqual(aggs, {r['uuid'] for r in rows})
        rows = self.resource_provider_aggregate_set(
            rp['uuid'], *[], generation=rp['generation'] + 1)
        self.assertEqual([], rows)

    def test_fail_incorrect_generation(self):
        rp = self.resource_provider_create()
        agg = str(uuid.uuid4())
        self.assertCommandFailed(
            "Please update the generation and try again.",
            self.resource_provider_aggregate_set,
            rp['uuid'], agg, generation=99999)

    def test_fail_generation_not_int(self):
        rp = self.resource_provider_create()
        agg = str(uuid.uuid4())
        self.assertCommandFailed(
            "invalid int value",
            self.resource_provider_aggregate_set,
            rp['uuid'], agg, generation='barney')

    def test_fail_if_incorrect_aggregate_uuid(self):
        rp = self.resource_provider_create()
        aggs = ['abc', 'efg']
        self.assertCommandFailed(
            "is not a 'uuid'",
            self.resource_provider_aggregate_set,
            rp['uuid'], *aggs, generation=rp['generation'])

    # In version 1.19 a generation is required.
    def test_fail_generation_arg_version_handling(self):
        rp = self.resource_provider_create()
        agg = str(uuid.uuid4())
        self.assertCommandFailed(
            "A generation must be specified.",
            self.resource_provider_aggregate_set,
            rp['uuid'], agg)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from osc_placement.tests.functional import base


def sorted_resources(resource):
    return ','.join(sorted(resource.split(',')))


class TestAllocationCandidate(base.BaseTestCase):
    VERSION = '1.10'

    def test_list_no_resource_specified_error(self):
        self.assertCommandFailed(
            'At least one --resource must be specified',
            self.openstack, 'allocation candidate list')

    def test_list_non_key_value_resource_specified_error(self):
        self.assertCommandFailed(
            'Arguments to --resource must be of form '
            '<resource_class>=<value>',
            self.openstack, 'allocation candidate list --resource VCPU')

    def test_list_empty(self):
        self.assertEqual([], self.allocation_candidate_list(
            resources=['MEMORY_MB=999999999']))

    def test_list_one(self):
        rp = self.resource_provider_create()
        self.resource_inventory_set(rp['uuid'], 'MEMORY_MB=1024')
        candidates = self.allocation_candidate_list(
            resources=('MEMORY_MB=256',))
        self.assertIn(
            rp['uuid'],
            [candidate['resource provider'] for candidate in candidates])

    def assertResourceEqual(self, r1, r2):
        self.assertEqual(sorted_resources(r1), sorted_resources(r2))

    def test_list_multiple(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()
        self.resource_inventory_set(
            rp1['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        self.resource_inventory_set(
            rp2['uuid'], 'MEMORY_MB=16384', 'DISK_GB=1024')
        candidates = self.allocation_candidate_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'))
        rps = {c['resource provider']: c for c in candidates}
        self.assertResourceEqual(
            'MEMORY_MB=1024,DISK_GB=80', rps[rp1['uuid']]['allocation'])
        self.assertResourceEqual(
            'MEMORY_MB=1024,DISK_GB=80', rps[rp2['uuid']]['allocation'])
        self.assertResourceEqual(
            'MEMORY_MB=0/8192,DISK_GB=0/512',
            rps[rp1['uuid']]['inventory used/capacity'])
        self.assertResourceEqual(
            'MEMORY_MB=0/16384,DISK_GB=0/1024',
            rps[rp2['uuid']]['inventory used/capacity'])

    def test_list_shared(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()
        self.resource_inventory_set(rp1['uuid'], 'MEMORY_MB=8192')
        self.resource_inventory_set(rp2['uuid'], 'DISK_GB=1024')
        agg = str(uuid.uuid4())
        self.resource_provider_aggregate_set(rp1['uuid'], agg)
        self.resource_provider_aggregate_set(rp2['uuid'], agg)
        self.resource_provider_trait_set(
            rp2['uuid'], 'MISC_SHARES_VIA_AGGREGATE')
        candidates = self.allocation_candidate_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'))
        rps = {c['resource provider']: c for c in candidates}
        self.assertResourceEqual(
            'MEMORY_MB=1024', rps[rp1['uuid']]['allocation'])
        self.assertResourceEqual(
            'DISK_GB=80', rps[rp2['uuid']]['allocation'])
        self.assertResourceEqual(
            'MEMORY_MB=0/8192', rps[rp1['uuid']]['inventory used/capacity'])
        self.assertResourceEqual(
            'DISK_GB=0/1024', rps[rp2['uuid']]['inventory used/capacity'])
        self.assertEqual(
            rps[rp2['uuid']]['#'], rps[rp1['uuid']]['#'])

    def test_fail_if_unknown_rc(self):
        self.assertCommandFailed(
            'No such resource',
            self.allocation_candidate_list,
            resources=('UNKNOWN=10',))


class TestAllocationCandidate112(TestAllocationCandidate):
    VERSION = '1.12'


class TestAllocationCandidate116(base.BaseTestCase):
    VERSION = '1.16'

    def test_list_limit(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()
        self.resource_inventory_set(
            rp1['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        self.resource_inventory_set(
            rp2['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')

        unlimited = self.allocation_candidate_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'))
        self.assertTrue(len(set([row['#'] for row in unlimited])) > 1)

        limited = self.allocation_candidate_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            limit=1)
        self.assertEqual(len(set([row['#'] for row in limited])), 1)


class TestAllocationCandidate117(base.BaseTestCase):
    VERSION = '1.17'

    # NOTE(cdent): The choice of traits here is important. We need to
    # make sure that we do not overlap with 'test_show_required_trait'
    # in TestResourceProvider118 which also creates some resource
    # providers. In a multi-process enviromment, the tests can race.
    def test_show_required_trait(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()
        self.resource_inventory_set(
            rp1['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        self.resource_inventory_set(
            rp2['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        self.resource_provider_trait_set(
            rp1['uuid'], 'STORAGE_DISK_SSD', 'HW_NIC_SRIOV')
        self.resource_provider_trait_set(
            rp2['uuid'], 'STORAGE_DISK_HDD', 'HW_NIC_SRIOV')

        rps = self.allocation_candidate_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=('STORAGE_DISK_SSD',))

        candidate_dict = {rp['resource provider']: rp for rp in rps}
        self.assertIn(rp1['uuid'], candidate_dict)
        self.assertNotIn(rp2['uuid'], candidate_dict)
        self.assertEqual(
            set(candidate_dict[rp1['uuid']]['traits'].split(',')),
            set(['STORAGE_DISK_SSD', 'HW_NIC_SRIOV']))

    # Prior to version 1.21 use the --aggregate-uuid arg should
    # be an errror.
    def test_fail_if_aggregate_uuid_wrong_version(self):
        self.assertCommandFailed(
            'Operation or argument is not supported with version 1.17',
            self.allocation_candidate_list,
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            aggregate_uuids=[str(uuid.uuid4())])
        # ...so as --member_of option
        self.assertCommandFailed(
            'Operation or argument is not supported with version 1.17',
            self.allocation_candidate_list,
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            member_of=[str(uuid.uuid4())])


class TestAllocationCandidate121(base.BaseTestCase):
    VERSION = '1.21'

    def test_return_properly_for_aggregate_uuid_request(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()
        self.resource_inventory_set(
            rp1['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        self.resource_inventory_set(
            rp2['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        agg = str(uuid.uuid4())

        self.resource_provider_aggregate_set(
            rp2['uuid'], agg, generation=1)

        # use --aggregate_uuids option
        rps, warning = self.allocation_candidate_list(
            resources=('MEMORY_MB=1024',),
            aggregate_uuids=[agg, str(uuid.uuid4())],
            may_print_to_stderr=True)

        candidate_dict = {rp['resource provider']: rp for rp in rps}
        self.assertEqual(1, len(candidate_dict))
        self.assertIn(rp2['uuid'], candidate_dict)
        self.assertNotIn(rp1['uuid'], candidate_dict)
        self.assertIn('The --aggregate-uuid option is deprecated, '
                      'please use --member-of instead.', warning)

        # validate --member_of option works as expected
        rps = self.allocation_candidate_list(
            resources=('MEMORY_MB=1024',),
            member_of=[agg])

        candidate_dict = {rp['resource provider']: rp for rp in rps}
        self.assertEqual(1, len(candidate_dict))
        self.assertIn(rp2['uuid'], candidate_dict)
        self.assertNotIn(rp1['uuid'], candidate_dict)

    # Specifying forbidden traits weren't available until version 1.22
    def test_fail_if_forbidden_trait_wrong_version(self):
        self.assertCommandFailed(
            'Operation or argument is not supported with version 1.21',
            self.allocation_candidate_list,
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            forbidden=('STORAGE_DISK_HDD',))


class TestAllocationCandidate122(base.BaseTestCase):
    VERSION = '1.22'

    def test_hide_forbidden_trait(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()
        rp3 = self.resource_provider_create()
        self.resource_inventory_set(
            rp1['uuid'], 'MEMORY_MB=1024', 'DISK_GB=256')
        self.resource_inventory_set(
            rp2['uuid'], 'MEMORY_MB=1024', 'DISK_GB=256')
        self.resource_inventory_set(
            rp3['uuid'], 'MEMORY_MB=1024', 'DISK_GB=256')
        self.resource_provider_trait_set(
            rp1['uuid'], 'STORAGE_DISK_SSD', 'HW_CPU_X86_BMI')
        self.resource_provider_trait_set(
            rp2['uuid'], 'STORAGE_DISK_HDD', 'HW_CPU_X86_BMI')
        self.resource_provider_trait_set(
            rp3['uuid'], 'STORAGE_DISK_HDD', 'HW_CPU_X86_BMI')

        rps = self.allocation_candidate_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=('HW_CPU_X86_BMI',),
            forbidden=('STORAGE_DISK_HDD',))

        self.assertEqual(1, len(rps))
        self.assertEqual(rp1['uuid'], rps[0]['resource provider'])

        rps = self.allocation_candidate_list(
            resources=('MEMORY_MB=1024', 'DISK_GB=80'),
            required=('HW_CPU_X86_BMI',),
            forbidden=('STORAGE_DISK_SSD',))

        uuids = [rp['resource provider'] for rp in rps]

        self.assertEqual(2, len(uuids))
        self.assertNotIn(rp1['uuid'], uuids)
        self.assertIn(rp2['uuid'], uuids)
        self.assertIn(rp3['uuid'], uuids)


class TestAllocationCandidate124(base.BaseTestCase):
    VERSION = '1.24'

    def test_member_of(self):
        rp1 = self.resource_provider_create()
        rp2 = self.resource_provider_create()
        self.resource_inventory_set(
            rp1['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        self.resource_inventory_set(
            rp2['uuid'], 'MEMORY_MB=8192', 'DISK_GB=512')
        agg1 = str(uuid.uuid4())
        agg2 = str(uuid.uuid4())
        agg3 = str(uuid.uuid4())

        self.resource_provider_aggregate_set(
            rp1['uuid'], agg1, agg3, generation=1)
        self.resource_provider_aggregate_set(
            rp2['uuid'], agg2, agg3, generation=1)

        agg1and3 = [agg1, agg3]
        agg1or3 = [agg1 + ',' + agg3]
        agg1or3_and_agg2 = [agg1 + ',' + agg3, agg2]

        rps = self.allocation_candidate_list(resources=('MEMORY_MB=1024',),
                                             member_of=agg1and3)
        candidate_dict = {rp['resource provider']: rp for rp in rps}
        self.assertEqual(1, len(candidate_dict))
        self.assertIn(rp1['uuid'], candidate_dict)

        rps = self.allocation_candidate_list(resources=('MEMORY_MB=1024',),
                                             member_of=agg1or3)
        candidate_dict = {rp['resource provider']: rp for rp in rps}
        self.assertEqual(2, len(candidate_dict))
        self.assertIn(rp1['uuid'], candidate_dict)
        self.assertIn(rp2['uuid'], candidate_dict)

        rps = self.allocation_candidate_list(resources=('MEMORY_MB=1024',),
                                             member_of=agg1or3_and_agg2)
        candidate_dict = {rp['resource provider']: rp for rp in rps}
        self.assertEqual(1, len(candidate_dict))
        self.assertIn(rp2['uuid'], candidate_dict)

    def test_fail_granular_wrong_version(self):
        groups = {'1': {'resources': ('VCPU=3',)}}
        self.assertCommandFailed(
            'Operation or argument is not supported with version 1.24',
            self.allocation_candidate_granular, groups=groups)


class TestAllocationCandidate129(base.BaseTestCase):
    VERSION = '1.29'

    def setUp(self):
        super(TestAllocationCandidate129, self).setUp()

        self.rp1 = self.resource_provider_create()
        self.rp1_1 = self.resource_provider_create(
            parent_provider_uuid=self.rp1['uuid'])
        self.rp1_2 = self.resource_provider_create(
            parent_provider_uuid=self.rp1['uuid'])

        self.agg1 = str(uuid.uuid4())
        self.agg2 = str(uuid.uuid4())
        self.resource_provider_aggregate_set(
            self.rp1_1['uuid'], self.agg1, generation=0)
        self.resource_provider_aggregate_set(
            self.rp1_2['uuid'], self.agg2, generation=0)

        self.resource_inventory_set(self.rp1['uuid'], 'DISK_GB=512')
        self.resource_inventory_set(
            self.rp1_1['uuid'], 'VCPU=8', 'MEMORY_MB=8192')
        self.resource_inventory_set(
            self.rp1_2['uuid'], 'VCPU=16', 'MEMORY_MB=8192')

        self.resource_provider_trait_set(self.rp1_1['uuid'], 'HW_CPU_X86_AVX')
        self.resource_provider_trait_set(self.rp1_2['uuid'], 'HW_CPU_X86_SSE')

    def test_granular_one_group(self):
        groups = {
            '1': {'resources': ('VCPU=3',)}
        }
        rows = self.allocation_candidate_granular(groups=groups)
        self.assertEqual(2, len(rows))

        numbers = {row['#'] for row in rows}
        self.assertEqual(2, len(numbers))

        rps = {row['resource provider'] for row in rows}
        self.assertEqual(2, len(rps))
        self.assertIn(self.rp1_1['uuid'], rps)
        self.assertIn(self.rp1_2['uuid'], rps)

    def test_granular_two_groups(self):
        groups = {
            '1': {'resources': ('VCPU=3',)},
            '2': {'resources': ('VCPU=3',)}
        }
        rows = self.allocation_candidate_granular(groups=groups)
        self.assertEqual(6, len(rows))

        numbers = {row['#'] for row in rows}
        self.assertEqual(4, len(numbers))

        rps = {row['resource provider'] for row in rows}
        self.assertEqual(2, len(rps))
        self.assertIn(self.rp1_1['uuid'], rps)
        self.assertIn(self.rp1_2['uuid'], rps)

        rows = self.allocation_candidate_granular(groups=groups,
                                                  group_policy='isolate')
        self.assertEqual(4, len(rows))

        numbers = {row['#'] for row in rows}
        self.assertEqual(2, len(numbers))

        rps = {row['resource provider'] for row in rows}
        self.assertEqual(2, len(rps))
        self.assertIn(self.rp1_1['uuid'], rps)
        self.assertIn(self.rp1_2['uuid'], rps)

        rows = self.allocation_candidate_granular(groups=groups,
                                                  group_policy='isolate',
                                                  limit=1)
        self.assertEqual(2, len(rows))

        numbers = {row['#'] for row in rows}
        self.assertEqual(1, len(numbers))

        rps = {row['resource provider'] for row in rows}
        self.assertEqual(2, len(rps))
        self.assertIn(self.rp1_1['uuid'], rps)
        self.assertIn(self.rp1_2['uuid'], rps)

    def test_granular_traits1(self):
        groups = {
            '1': {'resources': ('VCPU=6',)},
            '2': {'resources': ('VCPU=10',),
                  'required': ['HW_CPU_X86_AVX']}
        }
        rows = self.allocation_candidate_granular(groups=groups,
                                                  group_policy='isolate')
        self.assertEqual(0, len(rows))

    def test_granular_traits2(self):
        groups = {
            '1': {'resources': ('VCPU=6',)},
            '2': {'resources': ('VCPU=10',),
                  'required': ['HW_CPU_X86_SSE']}
        }
        rows = self.allocation_candidate_granular(groups=groups,
                                                  group_policy='isolate')
        self.assertEqual(2, len(rows))

        numbers = {row['#'] for row in rows}
        self.assertEqual(1, len(numbers))

        rps = {row['resource provider'] for row in rows}
        self.assertEqual(2, len(rps))
        self.assertIn(self.rp1_1['uuid'], rps)
        self.assertIn(self.rp1_2['uuid'], rps)

    def test_list_with_any_traits_old_microversion(self):
        groups = {
            '': {
                'resources': ('DISK_GB=1',),
                'required': ('STORAGE_DISK_HDD,STORAGE_DISK_SSD',),
            },
            '1': {
                'resources': ('VCPU=1',),
                'required': ('HW_CPU_X86_AVX',),
            }
        }
        self.assertCommandFailed(
            'Operation or argument is not supported with version 1.29',
            self.allocation_candidate_granular, groups=groups
        )


class TestAllocationCandidate139(base.BaseTestCase):
    VERSION = '1.39'

    def setUp(self):
        super(TestAllocationCandidate139, self).setUp()

        self.rp1 = self.resource_provider_create()
        self.rp1_1 = self.resource_provider_create(
            parent_provider_uuid=self.rp1['uuid'])
        self.rp1_2 = self.resource_provider_create(
            parent_provider_uuid=self.rp1['uuid'])

        self.resource_inventory_set(self.rp1['uuid'], 'DISK_GB=512')
        self.resource_inventory_set(
            self.rp1_1['uuid'], 'VCPU=8', 'MEMORY_MB=8192')
        self.resource_inventory_set(
            self.rp1_2['uuid'], 'VCPU=16', 'MEMORY_MB=8192')

        self.resource_provider_trait_set(self.rp1['uuid'], 'STORAGE_DISK_HDD')
        self.resource_provider_trait_set(self.rp1_1['uuid'], 'HW_CPU_X86_AVX')
        self.resource_provider_trait_set(self.rp1_2['uuid'], 'HW_CPU_X86_SSE')

        self.rp2 = self.resource_provider_create()
        self.rp2_1 = self.resource_provider_create(
            parent_provider_uuid=self.rp2['uuid'])
        self.rp2_2 = self.resource_provider_create(
            parent_provider_uuid=self.rp2['uuid'])

        self.resource_inventory_set(self.rp2['uuid'], 'DISK_GB=512')
        self.resource_inventory_set(
            self.rp2_1['uuid'], 'VCPU=8', 'MEMORY_MB=8192')
        self.resource_inventory_set(
            self.rp2_2['uuid'], 'VCPU=16', 'MEMORY_MB=8192')

        self.resource_provider_trait_set(self.rp2['uuid'], 'STORAGE_DISK_SSD')
        self.resource_provider_trait_set(self.rp2_1['uuid'], 'HW_CPU_X86_AVX')
        self.resource_provider_trait_set(self.rp2_2['uuid'], 'HW_CPU_X86_SSE')

    def test_list_with_any_traits(self):
        # asking for HDD and AVX that is only on the first tree as the second
        # has SSD instead
        groups = {
            '': {
                'resources': ('DISK_GB=1',),
                'required': ('STORAGE_DISK_HDD',),
            },
            '1': {
                'resources': ('VCPU=1',),
                'required': ('HW_CPU_X86_AVX',),
            }
        }
        rows = self.allocation_candidate_granular(groups=groups)

        # we expect one candidate
        numbers = {row['#'] for row in rows}
        self.assertEqual(1, len(numbers))
        # with two groups satisfied
        self.assertEqual(2, len(rows))

        rps = {row['resource provider'] for row in rows}
        self.assertEqual({self.rp1['uuid'], self.rp1_1['uuid']}, rps)

        # extend this by asking for SSD or HDD
        groups = {
            '': {
                'resources': ('DISK_GB=1',),
                'required': ('STORAGE_DISK_HDD,STORAGE_DISK_SSD',),
            },
            '1': {
                'resources': ('VCPU=1',),
                'required': ('HW_CPU_X86_AVX',),
            }
        }
        rows = self.allocation_candidate_granular(groups=groups)

        # we expect two candidates now as both tree matches
        numbers = {row['#'] for row in rows}
        self.assertEqual(2, len(numbers))
        # with two groups satisfied each
        self.assertEqual(4, len(rows))

        rps = {row['resource provider'] for row in rows}
        self.assertEqual(
            {
                self.rp1['uuid'], self.rp1_1['uuid'],
                self.rp2['uuid'], self.rp2_1['uuid'],
            },
            rps
        )

        # make it crazy by asking for (HDD or SSD) and SSD and not HDD
        # this basically means SSD but tests all the branches of the client
        # code
        # similarly for the granular group ask for (AVX or SSE) and not SSE
        groups = {
            '': {
                'resources': ('DISK_GB=1',),
                'required': (
                    'STORAGE_DISK_HDD,STORAGE_DISK_SSD',
                    'STORAGE_DISK_SSD'
                ),
                'forbidden': ('STORAGE_DISK_HDD',),
            },
            '1': {
                'resources': ('VCPU=1',),
                'required': ('HW_CPU_X86_AVX,HW_CPU_X86_SSE',),
                'forbidden': ('HW_CPU_X86_SSE',),
            }
        }
        rows = self.allocation_candidate_granular(groups=groups)

        # SSD and AVX means we only the second tree matches with a single
        # candidate
        numbers = {row['#'] for row in rows}
        self.assertEqual(1, len(numbers))
        # with two groups satisfied
        self.assertEqual(2, len(rows))

        rps = {row['resource provider'] for row in rows}
        self.assertEqual({self.rp2['uuid'], self.rp2_1['uuid']}, rps)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from distutils.version import StrictVersion
import operator


NEGOTIATE_VERSIONS = [
    '1',  # Added for auto choice for appropriate version
]
SUPPORTED_MICROVERSIONS = [
    '1.0',
    '1.1',
    '1.2',
    '1.3',
    '1.4',
    '1.5',
    '1.6',
    '1.7',
    '1.8',
    '1.9',
    '1.10',
    '1.11',
    '1.12',
    '1.13',  # unused
    '1.14',
    '1.15',  # unused
    '1.16',
    '1.17',
    '1.18',
    '1.19',
    '1.20',  # unused
    '1.21',
    '1.22',
    '1.23',  # unused
    '1.24',
    '1.25',
    '1.26',  # unused
    '1.27',  # unused
    '1.28',  # Added for provider allocation (un)set (Ussuri)
    '1.29',
    '1.37',  # unused
    '1.38',  # Added for consumer types (Xena)
    '1.39',  # Added any-traits support (Yoga)
]
SUPPORTED_VERSIONS = SUPPORTED_MICROVERSIONS + NEGOTIATE_VERSIONS
# The max microversion lower than which are all supported by this client.
# This is used to automatically pick up the microversion to use. Change this
# when you add a microversion to the `_SUPPORTED_VERSIONS` without a gap.
# TestVersion.test_max_version_consistency checks its consistency.
MAX_VERSION_NO_GAP = '1.29'


def _op(func, b, msg):
    return lambda a: func(StrictVersion(a), StrictVersion(b)) or msg


def lt(b):
    msg = 'requires version less than %s' % b
    return _op(operator.lt, b, msg)


def le(b):
    msg = 'requires at most version %s' % b
    return _op(operator.le, b, msg)


def eq(b):
    msg = 'requires version %s' % b
    return _op(operator.eq, b, msg)


def ne(b):
    msg = 'can not use version %s' % b
    return _op(operator.ne, b, msg)


def ge(b):
    msg = 'requires at least version %s' % b
    return _op(operator.ge, b, msg)


def gt(b):
    msg = 'requires version greater than %s' % b
    return _op(operator.gt, b, msg)


def _compare(ver, *predicates, **kwargs):
    func = kwargs.get('op', all)
    if func(p(ver) is True for p in predicates):
        return True
    # construct an error message if the requirement not satisfied
    err_msg = 'Operation or argument is not supported with version %s; ' % ver
    err_detail = [p(ver) for p in predicates if p(ver) is not True]
    logic = ', and ' if func is all else ', or '
    return err_msg + logic.join(err_detail)


def compare(ver, *predicates, **kwargs):
    """Validate version satisfies provided predicates.

    kwargs['exc'] - boolean whether exception should be raised
    kwargs['op'] - (all, any) how predicates should be checked

    Examples:
        compare('1.1', version.gt('1.2'), exc=False) - False
        compare('1.1', version.eq('1.0'), version.eq('1.1'), op=any) - True

    """
    exc = kwargs.get('exc', True)
    result = _compare(ver, *predicates, **kwargs)
    if result is not True:
        if exc:
            raise ValueError(result)
        return False
    return True


def check(*predicates, **check_kwargs):
    """Decorator for command object method.

    See `compare`

    """
    def wrapped(func):
        def inner(self, *args, **kwargs):
            compare(get_version(self), *predicates, **check_kwargs)
            return func(self, *args, **kwargs)
        return inner
    return wrapped


def get_version(obj):
    """Extract version from a command object."""
    try:
        if obj.app.client_manager.session is None:
            return MAX_VERSION_NO_GAP
        version = obj.app.client_manager.placement.api_version
    except AttributeError:
        # resource does not have api_version attr when docs are generated
        # so let's use the minimal one
        version = SUPPORTED_VERSIONS[0]
    return version


class CheckerMixin(object):
    def check_version(self, *predicates, **kwargs):
        return compare(get_version(self), *predicates, **kwargs)

    def compare_version(self, *predicates, **kwargs):
        return compare(get_version(self), *predicates, exc=False, **kwargs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import logging
import random

import fixtures

from openstackclient import shell
from oslotest import base
from placement.tests.functional.fixtures import capture
from placement.tests.functional.fixtures import placement
import simplejson as json


# A list of logger names that will be reset to a log level
# of WARNING. Due (we think) to a poor interaction between the
# way osc does logging and oslo.logging, all packages are producing
# DEBUG logs. This results in test attachments (when capturing logs)
# that are sometimes larger than subunit.parser can deal with. The
# packages chosen here are ones that do not provide useful information.
RESET_LOGGING = [
    'keystoneauth.session',
    'oslo_policy.policy',
    'placement.objects.trait',
    'placement.objects.resource_class',
    'placement.objects.resource_provider',
    'oslo_concurrency.lockutils',
    'osc_lib.shell',
]

RP_PREFIX = 'osc-placement-functional-tests-'

ARGUMENTS_MISSING = 'the following arguments are required'
ARGUMENTS_REQUIRED = 'the following arguments are required: %s'


class CommandException(Exception):
    def __init__(self, *args, **kwargs):
        super(CommandException, self).__init__(args[0])
        self.cmd = kwargs['cmd']


class BaseTestCase(base.BaseTestCase):
    VERSION = '1.0'

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.useFixture(capture.Logging())
        self.placement = self.useFixture(placement.PlacementFixture())

        # Work around needing to reset the session's notion of where
        # we are going.
        def mock_get(obj, instance, owner):
            return obj.factory(instance)

        # NOTE(cdent): This is fragile, but is necessary to work around
        # the rather complex start up optimizations that are done in osc_lib.
        # If/when osc_lib changes this will at least fail fast.
        self.useFixture(fixtures.MonkeyPatch(
            'osc_lib.clientmanager.ClientCache.__get__',
            mock_get))

        # Reset log level on a set of packages. See comment on RESET_LOGGING
        # assigment, above.
        for name in RESET_LOGGING:
            logging.getLogger(name).setLevel(logging.WARNING)

    def openstack(self, cmd, may_fail=False, use_json=False,
                  may_print_to_stderr=False):
        to_exec = []
        # Make all requests as a noauth admin user.
        to_exec += [
            '--os-endpoint', self.placement.endpoint,
            '--os-token', self.placement.token,
            '--os-auth-type', 'admin_token',
        ]
        if self.VERSION is not None:
            to_exec += ['--os-placement-api-version', self.VERSION]
        to_exec += cmd.split()
        if use_json:
            to_exec += ['-f', 'json']

        # Context manager here instead of setUp because we only want
        # output trapping around the run().
        self.output = io.StringIO()
        self.error = io.StringIO()
        stdout_fix = fixtures.MonkeyPatch('sys.stdout', self.output)
        stderr_fix = fixtures.MonkeyPatch('sys.stderr', self.error)
        with stdout_fix, stderr_fix:
            try:
                os_shell = shell.OpenStackShell()
                return_code = os_shell.run(to_exec)
            # Catch SystemExit to trap some error responses, mostly from the
            # argparse lib which has a tendency to exit for you instead of
            # politely telling you it wants to.
            except SystemExit as exc:
                return_code = exc.code

        # We may have error/warning messages in stderr, so treat it
        # separately from the stdout.
        output = self.output.getvalue()
        error = self.error.getvalue()

        if return_code:
            msg = 'Command: "%s"\noutput: %s' % (' '.join(to_exec), error)
            if not may_fail:
                raise CommandException(msg, cmd=' '.join(to_exec))

        if use_json and output:
            output = json.loads(output)

        if may_print_to_stderr:
            return output, error

        if error:
            msg = ('Test code error - The command did not fail but it '
                   'has a warning message. Set the "may_print_to_stderr" '
                   'argument to true to get and validate the message:\n'
                   'Command: "%s"\nstderr: %s') % (
                ' '.join(to_exec), error)
            raise CommandException(msg, cmd=' '.join(to_exec))
        return output

    def rand_name(self, name='', prefix=None):
        """Generate a random name that includes a random number

        :param str name: The name that you want to include
        :param str prefix: The prefix that you want to include
        :return: a random name. The format is
                 '<prefix>-<name>-<random number>'.
                 (e.g. 'prefixfoo-namebar-154876201')
        :rtype: string
        """
        # NOTE(lajos katona): This method originally is in tempest-lib.
        randbits = str(random.randint(1, 0x7fffffff))
        rand_name = randbits
        if name:
            rand_name = name + '-' + rand_name
        if prefix:
            rand_name = prefix + '-' + rand_name
        return rand_name

    def assertCommandFailed(self, message, func, *args, **kwargs):
        signature = [func]
        signature.extend(args)
        try:
            func(*args, **kwargs)
            self.fail('Command does not fail as required (%s)' % signature)
        except CommandException as e:
            self.assertIn(
                message, str(e),
                'Command "%s" fails with different message' % e.cmd)

    def resource_provider_create(self,
                                 name='',
                                 parent_provider_uuid=None):
        if not name:
            name = self.rand_name(name='', prefix=RP_PREFIX)

        to_exec = 'resource provider create ' + name
        if parent_provider_uuid is not None:
            to_exec += ' --parent-provider ' + parent_provider_uuid
        res = self.openstack(to_exec, use_json=True)

        def cleanup():
            try:
                self.resource_provider_delete(res['uuid'])
            except CommandException as exc:
                # may have already been deleted by a test case
                err_message = str(exc).lower()
                if 'no resource provider' not in err_message:
                    raise
        self.addCleanup(cleanup)

        return res

    def resource_provider_set(self, uuid, name, parent_provider_uuid=None):
        to_exec = 'resource provider set ' + uuid + ' --name ' + name
        if parent_provider_uuid is not None:
            to_exec += ' --parent-provider ' + parent_provider_uuid
        return self.openstack(to_exec, use_json=True)

    def resource_provider_show(self, uuid, allocations=False):
        cmd = 'resource provider show ' + uuid
        if allocations:
            cmd = cmd + ' --allocations'

        return self.openstack(cmd, use_json=True)

    def resource_provider_list(self, uuid=None, name=None,
                               aggregate_uuids=None, resources=None,
                               in_tree=None, required=None, forbidden=None,
                               member_of=None, may_print_to_stderr=False):
        to_exec = 'resource provider list'
        if uuid:
            to_exec += ' --uuid ' + uuid
        if name:
            to_exec += ' --name ' + name
        if aggregate_uuids:
            to_exec += ' ' + ' '.join(
                '--aggregate-uuid %s' % a for a in aggregate_uuids)
        if resources:
            to_exec += ' ' + ' '.join('--resource %s' % r for r in resources)
        if in_tree:
            to_exec += ' --in-tree ' + in_tree
        if required:
            to_exec += ' ' + ' '.join('--required %s' % t for t in required)
        if forbidden:
            to_exec += ' ' + ' '.join('--forbidden %s' % f for f in forbidden)
        if member_of:
            to_exec += ' ' + ' '.join(
                ['--member-of %s' % m for m in member_of])

        return self.openstack(
            to_exec, use_json=True, may_print_to_stderr=may_print_to_stderr)

    def resource_provider_delete(self, uuid):
        return self.openstack('resource provider delete ' + uuid)

    def resource_allocation_show(self, consumer_uuid, columns=()):
        cmd = 'resource provider allocation show ' + consumer_uuid
        cmd += ' '.join(' --column %s' % c for c in columns)
        return self.openstack(cmd, use_json=True)

    def resource_allocation_set(self, consumer_uuid, allocations,
                                project_id=None, user_id=None,
                                consumer_type=None, use_json=True,
                                may_print_to_stderr=False):
        cmd = 'resource provider allocation set {allocs} {uuid}'.format(
            uuid=consumer_uuid,
            allocs=' '.join('--allocation {}'.format(a) for a in allocations)
        )
        if project_id:
            cmd += ' --project-id %s' % project_id
        if user_id:
            cmd += ' --user-id %s' % user_id
        if consumer_type:
            cmd += ' --consumer-type %s' % consumer_type
        result = self.openstack(
            cmd, use_json=use_json, may_print_to_stderr=may_print_to_stderr)

        def cleanup(uuid):
            try:
                self.openstack('resource provider allocation delete ' + uuid)
            except CommandException as exc:
                # may have already been deleted by a test case
                if 'not found' in str(exc).lower():
                    pass
        self.addCleanup(cleanup, consumer_uuid)

        return result

    def resource_allocation_unset(
        self, consumer_uuid, provider=None, resource_class=None, use_json=True,
        columns=(),
    ):
        cmd = 'resource provider allocation unset %s' % consumer_uuid
        if resource_class:
            cmd += ' ' + ' '.join(
                '--resource-class %s' % rc for rc in resource_class)
        if provider:
            # --provider can be specified multiple times so if we only get
            # a single string value convert to a list.
            if isinstance(provider, str):
                provider = [provider]
            cmd += ' ' + ' '.join(
                '--provider %s' % rp_uuid for rp_uuid in provider)

        cmd += ' '.join(' --column %s' % c for c in columns)

        result = self.openstack(cmd, use_json=use_json)

        def cleanup(uuid):
            try:
                self.openstack('resource provider allocation delete ' + uuid)
            except CommandException as exc:
                # may have already been deleted by a test case
                if 'not found' in str(exc).lower():
                    pass
        self.addCleanup(cleanup, consumer_uuid)

        return result

    def resource_allocation_delete(self, consumer_uuid):
        cmd = 'resource provider allocation delete ' + consumer_uuid
        return self.openstack(cmd)

    def resource_inventory_show(
        self, uuid, resource_class, *, include_used=False,
    ):
        resource = self.openstack(
            f'resource provider inventory show {uuid} {resource_class}',
            use_json=True,
        )
        if not include_used:
            del resource['used']

        return resource

    def resource_inventory_list(self, uuid, *, include_used=False):
        resources = self.openstack(
            f'resource provider inventory list {uuid}',
            use_json=True,
        )
        if not include_used:
            for resource in resources:
                del resource['used']

        return resources

    def resource_inventory_delete(self, uuid, resource_class=None):
        cmd = 'resource provider inventory delete {uuid}'.format(uuid=uuid)
        if resource_class:
            cmd += ' --resource-class ' + resource_class
        self.openstack(cmd)

    def resource_inventory_set(self, uuid, *resources, **kwargs):
        opts = []
        if kwargs.get('aggregate'):
            opts.append('--aggregate')
        if kwargs.get('amend'):
            opts.append('--amend')
        if kwargs.get('dry_run'):
            opts.append('--dry-run')
        fmt = 'resource provider inventory set {uuid} {resources} {opts}'
        cmd = fmt.format(
            uuid=uuid,
            resources=' '.join(['--resource %s' % r for r in resources]),
            opts=' '.join(opts))
        return self.openstack(cmd, use_json=True)

    def resource_inventory_class_set(self, uuid, resource_class, **kwargs):
        opts = ['--%s=%s' % (k, v) for k, v in kwargs.items()]
        cmd = 'resource provider inventory class set {uuid} {rc} {opts}'.\
            format(uuid=uuid, rc=resource_class, opts=' '.join(opts))
        return self.openstack(cmd, use_json=True)

    def resource_provider_show_usage(self, uuid):
        return self.openstack('resource provider usage show ' + uuid,
                              use_json=True)

    def resource_show_usage(self, project_id, user_id=None):
        cmd = 'resource usage show %s' % project_id
        if user_id:
            cmd += ' --user-id %s' % user_id
        return self.openstack(cmd, use_json=True)

    def resource_provider_aggregate_list(self, uuid):
        return self.openstack('resource provider aggregate list ' + uuid,
                              use_json=True)

    def resource_provider_aggregate_set(self, uuid, *aggregates,
                                        **kwargs):
        generation = kwargs.get('generation')
        cmd = 'resource provider aggregate set %s ' % uuid
        cmd += ' '.join('--aggregate %s' % aggregate
                        for aggregate in aggregates)
        if generation is not None:
            cmd += ' --generation %s' % generation
        return self.openstack(cmd, use_json=True)

    def resource_class_list(self):
        return self.openstack('resource class list', use_json=True)

    def resource_class_show(self, name):
        return self.openstack('resource class show ' + name, use_json=True)

    def resource_class_create(self, name):
        return self.openstack('resource class create ' + name)

    def resource_class_set(self, name):
        return self.openstack('resource class set ' + name)

    def resource_class_delete(self, name):
        return self.openstack('resource class delete ' + name)

    def trait_list(self, name=None, associated=False):
        cmd = 'trait list'
        if name:
            cmd += ' --name ' + name
        if associated:
            cmd += ' --associated'
        return self.openstack(cmd, use_json=True)

    def trait_show(self, name):
        cmd = 'trait show %s' % name
        return self.openstack(cmd, use_json=True)

    def trait_create(self, name):
        cmd = 'trait create %s' % name
        self.openstack(cmd)

        def cleanup():
            try:
                self.trait_delete(name)
            except CommandException as exc:
                # may have already been deleted by a test case
                err_message = str(exc).lower()
                if 'http 404' not in err_message:
                    raise
        self.addCleanup(cleanup)

    def trait_delete(self, name):
        cmd = 'trait delete %s' % name
        self.openstack(cmd)

    def resource_provider_trait_list(self, uuid):
        cmd = 'resource provider trait list %s ' % uuid
        return self.openstack(cmd, use_json=True)

    def resource_provider_trait_set(self, uuid, *traits):
        cmd = 'resource provider trait set %s ' % uuid
        cmd += ' '.join('--trait %s' % trait for trait in traits)
        return self.openstack(cmd, use_json=True)

    def resource_provider_trait_delete(self, uuid):
        cmd = 'resource provider trait delete %s ' % uuid
        self.openstack(cmd)

    def allocation_candidate_list(self, resources=None, required=None,
                                  forbidden=None, limit=None,
                                  aggregate_uuids=None, member_of=None,
                                  may_print_to_stderr=False):
        cmd = 'allocation candidate list '
        cmd += self._allocation_candidates_option(
            resources, required, forbidden, aggregate_uuids, member_of)
        if limit is not None:
            cmd += ' --limit %d' % limit
        return self.openstack(
            cmd, use_json=True, may_print_to_stderr=may_print_to_stderr)

    def allocation_candidate_granular(self, groups, group_policy=None,
                                      limit=None):
        cmd = 'allocation candidate list '
        for suffix, req_group in groups.items():
            if suffix:
                cmd += ' --group %s' % suffix
            cmd += self._allocation_candidates_option(**req_group)
            if limit is not None:
                cmd += ' --limit %d' % limit
            if group_policy is not None:
                cmd += ' --group-policy %s' % group_policy
        return self.openstack(cmd, use_json=True)

    def _allocation_candidates_option(self, resources=None, required=None,
                                      forbidden=None, aggregate_uuids=None,
                                      member_of=None):
        opt = ''
        if resources:
            opt += ' ' + ' '.join(
                '--resource %s' % resource for resource in resources)
        if required is not None:
            opt += ''.join([' --required %s' % t for t in required])
        if forbidden:
            opt += ' ' + ' '.join('--forbidden %s' % f for f in forbidden)
        if aggregate_uuids:
            opt += ' ' + ' '.join(
                '--aggregate-uuid %s' % a for a in aggregate_uuids)
        if member_of:
            opt += ' ' + ' '.join(['--member-of %s' % m for m in member_of])
        return opt



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from osc_placement import version


BASE_URL = '/allocations'


def parse_allocations(allocation_strings):
    allocations = {}
    for allocation_string in allocation_strings:
        if '=' not in allocation_string or ',' not in allocation_string:
            raise ValueError('Incorrect allocation string format')
        parsed = dict(kv.split('=') for kv in allocation_string.split(','))
        if 'rp' not in parsed:
            raise ValueError('Resource provider parameter is required '
                             'for allocation string')
        resources = {k: int(v) for k, v in parsed.items() if k != 'rp'}
        if parsed['rp'] not in allocations:
            allocations[parsed['rp']] = resources
        else:
            prev_rp = allocations[parsed['rp']]
            for resource, value in resources.items():
                if resource in prev_rp and prev_rp[resource] != value:
                    raise exceptions.CommandError(
                        'Conflict detected for '
                        'resource provider {} resource class {}'.format(
                            parsed['rp'], resource))
            allocations[parsed['rp']].update(resources)
    return allocations


class SetAllocation(command.Lister, version.CheckerMixin):
    """Replaces the set of resource allocation(s) for a given consumer.

    Note that this is a full replacement of the existing allocations. If you
    want to retain the existing allocations and add a new resource class
    allocation, you must specify all resource class allocations, old and new.

    From ``--os-placement-api-version 1.8`` it is required to specify
    ``--project-id`` and ``--user-id`` to set allocations. It is highly
    recommended to provide a ``--project-id`` and ``--user-id`` when setting
    allocations for accounting and data consistency reasons.

    Starting with ``--os-placement-api-version 1.12`` the API response
    contains the ``project_id`` and ``user_id`` of allocations which also
    appears in the CLI output.

    Starting with ``--os-placement-api-version 1.28`` a consumer generation is
    used which facilitates safe concurrent modification of an allocation.

    Starting with ``--os-placement-api-version 1.38`` it is required to specify
    ``--consumer-type`` to set allocations. It is helpful to provide a
    ``--consumer-type`` when setting allocations so that resource usages can be
    filtered on consumer types.
    """

    def get_parser(self, prog_name):
        parser = super(SetAllocation, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the consumer'
        )
        parser.add_argument(
            '--allocation',
            metavar='<rp=resource-provider-id,'
                    'resource-class-name=amount-of-resource-used>',
            action='append',
            default=[],
            help='Create (or update) an allocation of a resource class. '
                 'Specify option multiple times to set multiple allocations.'
        )
        parser.add_argument(
            '--project-id',
            metavar='project_id',
            help='ID of the consuming project. '
                 'This option is required starting from '
                 '``--os-placement-api-version 1.8``.',
            required=self.compare_version(version.ge('1.8'))
        )
        parser.add_argument(
            '--user-id',
            metavar='user_id',
            help='ID of the consuming user. '
                 'This option is required starting from '
                 '``--os-placement-api-version 1.8``.',
            required=self.compare_version(version.ge('1.8'))
        )
        parser.add_argument(
            '--consumer-type',
            metavar='consumer_type',
            help='The type of the consumer. '
                 'This option is required starting from '
                 '``--os-placement-api-version 1.38``.',
            required=self.compare_version(version.ge('1.38'))
        )
        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement
        url = BASE_URL + '/' + parsed_args.uuid

        # Determine if we need to honor consumer generations.
        supports_consumer_generation = self.compare_version(version.ge('1.28'))
        if supports_consumer_generation:
            # Get the existing consumer generation via GET.
            payload = http.request('GET', url).json()
            consumer_generation = payload.get('consumer_generation')

        allocations = parse_allocations(parsed_args.allocation)
        if not allocations:
            raise exceptions.CommandError(
                'At least one resource allocation must be specified')

        if self.compare_version(version.ge('1.12')):
            allocations = {
                rp: {'resources': resources}
                for rp, resources in allocations.items()}
        else:
            allocations = [
                {'resource_provider': {'uuid': rp}, 'resources': resources}
                for rp, resources in allocations.items()]

        payload = {'allocations': allocations}
        # Include consumer_generation for 1.28+. Note that if this is the
        # first set of allocations the consumer_generation will be None.
        if supports_consumer_generation:
            payload['consumer_generation'] = consumer_generation
        if self.compare_version(version.ge('1.8')):
            payload['project_id'] = parsed_args.project_id
            payload['user_id'] = parsed_args.user_id
        elif parsed_args.project_id or parsed_args.user_id:
            self.log.warning('--project-id and --user-id options do not '
                             'affect allocation for '
                             '--os-placement-api-version less than 1.8')
        if self.compare_version(version.ge('1.38')):
            payload['consumer_type'] = parsed_args.consumer_type
        elif parsed_args.consumer_type:
            self.log.warning('--consumer-type option does not affect '
                             'allocation for --os-placement-api-version less '
                             'than 1.38')
        http.request('PUT', url, json=payload)
        resp = http.request('GET', url).json()
        per_provider = resp['allocations'].items()
        props = {}

        fields = ('resource_provider', 'generation', 'resources')
        if self.compare_version(version.ge('1.12')):
            fields += ('project_id', 'user_id')
            props['project_id'] = resp['project_id']
            props['user_id'] = resp['user_id']
        if self.compare_version(version.ge('1.38')):
            fields += ('consumer_type',)
            props['consumer_type'] = resp['consumer_type']

        allocs = [dict(resource_provider=k, **props, **v)
                  for k, v in per_provider]
        rows = (utils.get_dict_properties(a, fields) for a in allocs)
        return fields, rows


class UnsetAllocation(command.Lister, version.CheckerMixin):
    """Removes one or more sets of provider allocations for a consumer.

    Note that omitting both the ``--provider`` and the ``--resource-class``
    option is equivalent to removing all allocations for the given consumer.

    This command requires ``--os-placement-api-version 1.12`` or greater. Use
    ``openstack resource provider allocation set`` for older versions.
    """

    def get_parser(self, prog_name):
        parser = super(UnsetAllocation, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<consumer_uuid>',
            help='UUID of the consumer. It is strongly recommended to use '
                 '``--os-placement-api-version 1.28`` or greater when using '
                 'this option to ensure the other allocation information is '
                 'retained. '
        )
        parser.add_argument(
            '--provider',
            metavar='provider_uuid',
            action='append',
            default=[],
            help='UUID of a specific resource provider from which to remove '
                 'allocations for the given consumer. This is useful when the '
                 'consumer has allocations on more than one provider, for '
                 'example after evacuating a server to another compute node '
                 'and you want to cleanup allocations on the source compute '
                 'node resource provider in order to delete it. Specify '
                 'multiple times to remove allocations against multiple '
                 'resource providers. Omit this option to remove all '
                 'allocations for the consumer, or to remove all allocations'
                 'of a specific resource class from all the resource provider '
                 'with the ``--resource_class`` option. '
        )
        parser.add_argument(
            '--resource-class',
            metavar='resource_class',
            action='append',
            default=[],
            help='Name of a resource class from which to remove allocations '
                 'for the given consumer. This is useful when the consumer '
                 'has allocations on more than one resource class. '
                 'By default, this will remove allocations for the given '
                 'resource class from all the providers. If ``--provider`` '
                 'option is also specified, allocations to remove will be '
                 'limited to that resource class of the given resource '
                 'provider.'
        )
        return parser

    # NOTE(mriedem): We require >= 1.12 because PUT requires project_id/user_id
    # since 1.8 but GET does not return project_id/user_id until 1.12 and we
    # do not want to add --project-id and --user-id options to this command
    # like in the set command. If someone needs to use an older microversion or
    # change the user/project they can use the set command.
    @version.check(version.ge('1.12'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement
        url = BASE_URL + '/' + parsed_args.uuid

        # Get the current allocations.
        payload = http.request('GET', url).json()
        allocations = payload['allocations']

        if parsed_args.resource_class:
            # Remove the given resource class. Do not error out if the
            # consumer does not have allocations against that resource
            # class.
            rp_uuids = set(allocations)
            if parsed_args.provider:
                # If providers are also specified, we limit to remove
                # allocations only from those providers
                rp_uuids &= set(parsed_args.provider)
            for rp_uuid in rp_uuids:
                for rc in parsed_args.resource_class:
                    allocations[rp_uuid]['resources'].pop(rc, None)
                if not allocations[rp_uuid]['resources']:
                    allocations.pop(rp_uuid, None)
        else:
            if parsed_args.provider:
                # Remove the given provider(s) from the allocations if it
                # exists. Do not error out if the consumer does not have
                # allocations against a provider in case we lost a race since
                # the allocations are in the state the user wants them in
                # anyway.
                for rp_uuid in parsed_args.provider:
                    allocations.pop(rp_uuid, None)
            else:
                # No --provider(s) specified so remove allocations from all
                # providers.
                allocations = {}

        supports_consumer_generation = self.compare_version(version.ge('1.28'))
        # 1.28+ allows PUTing an empty allocations dict as long as a
        # consumer_generation is specified.
        if allocations or supports_consumer_generation:
            payload['allocations'] = allocations
            http.request('PUT', url, json=payload)
        else:
            # The user must have removed all of the allocations so just DELETE
            # the allocations since we cannot PUT with an empty allocations
            # dict before 1.28.
            http.request('DELETE', url)

        resp = http.request('GET', url).json()
        per_provider = resp['allocations'].items()
        props = {}

        fields = ('resource_provider', 'generation', 'resources',
                  'project_id', 'user_id')
        if self.compare_version(version.ge('1.38')):
            fields += ('consumer_type',)
            props['consumer_type'] = resp.get('consumer_type')
        allocs = [dict(project_id=resp['project_id'], user_id=resp['user_id'],
                       resource_provider=k, **props, **v)
                  for k, v in per_provider]
        rows = (utils.get_dict_properties(a, fields) for a in allocs)
        return fields, rows


class ShowAllocation(command.Lister, version.CheckerMixin):
    """Show resource allocations for a given consumer.

    Starting with ``--os-placement-api-version 1.12`` the API response contains
    the ``project_id`` and ``user_id`` of allocations which also appears in the
    CLI output.

    Starting with ``--os-placement-api-version 1.38`` the API response contains
    the ``consumer_type`` of consumer which also appears in the CLI output.
    """

    def get_parser(self, prog_name):
        parser = super(ShowAllocation, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the consumer'
        )

        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = BASE_URL + '/' + parsed_args.uuid
        resp = http.request('GET', url).json()
        per_provider = resp['allocations'].items()
        props = {}

        fields = ('resource_provider', 'generation', 'resources')
        if self.compare_version(version.ge('1.12')):
            fields += ('project_id', 'user_id')
            props['project_id'] = resp.get('project_id')
            props['user_id'] = resp.get('user_id')
        if self.compare_version(version.ge('1.38')):
            fields += ('consumer_type',)
            props['consumer_type'] = resp.get('consumer_type')

        allocs = [dict(resource_provider=k, **props, **v)
                  for k, v in per_provider]

        rows = (utils.get_dict_properties(a, fields) for a in allocs)
        return fields, rows


class DeleteAllocation(command.Command):
    """Delete all resource allocations for a given consumer."""

    def get_parser(self, prog_name):
        parser = super(DeleteAllocation, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the consumer'
        )

        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = BASE_URL + '/' + parsed_args.uuid
        http.request('DELETE', url)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from unittest import mock

from oslotest import base

import osc_placement.plugin as plugin


class TestPlugin(base.BaseTestCase):
    def test_build_option_parser(self):
        parser = plugin.build_option_parser(argparse.ArgumentParser())

        args = parser.parse_args(['--os-placement-api-version=1.0'])
        self.assertEqual('1.0', args.os_placement_api_version)
        args = parser.parse_args(['--os-placement-api-version', '1.0'])
        self.assertEqual('1.0', args.os_placement_api_version)

    @mock.patch('osc_placement.http.SessionClient')
    def test_make_client(self, mock_session_client):
        instance = mock.Mock(_api_version={'placement': '1.0'})

        plugin.make_client(instance)
        mock_session_client.assert_called_with(
            session=instance.session,
            ks_filter={
                'service_type': 'placement',
                'region_name': instance._region_name,
                'interface': instance.interface
            },
            api_version='1.0'
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from osc_lib import exceptions
from oslotest import base

from osc_placement.resources import allocation


class TestAllocation(base.BaseTestCase):
    def test_parse_allocations(self):
        rp1 = str(uuid.uuid4())
        rp2 = str(uuid.uuid4())
        allocations = [
            'rp={},VCPU=4,MEMORY_MB=16324'.format(rp1),
            'rp={},VCPU=4,DISK_GB=4096'.format(rp2)]
        expected = {
            rp1: {'VCPU': 4, 'MEMORY_MB': 16324},
            rp2: {'VCPU': 4, 'DISK_GB': 4096},
        }
        self.assertDictEqual(
            expected, allocation.parse_allocations(allocations))

    def test_merge_allocations(self):
        rp1 = str(uuid.uuid4())
        allocations = [
            'rp={},VCPU=4,MEMORY_MB=16324'.format(rp1),
            'rp={},VCPU=4,DISK_GB=4096'.format(rp1)]
        expected = {
            rp1: {'VCPU': 4, 'MEMORY_MB': 16324, 'DISK_GB': 4096}}
        self.assertEqual(expected, allocation.parse_allocations(allocations))

    def test_fail_if_cannot_merge_allocations(self):
        rp1 = str(uuid.uuid4())
        allocations = [
            'rp={},VCPU=4,MEMORY_MB=16324'.format(rp1),
            'rp={},VCPU=8,DISK_GB=4096'.format(rp1)]
        ex = self.assertRaises(
            exceptions.CommandError, allocation.parse_allocations, allocations)
        self.assertEqual(
            'Conflict detected for resource provider %s resource class VCPU' %
            rp1, str(ex))

    def test_fail_if_incorrect_format(self):
        allocations = ['incorrect_format']
        self.assertRaisesRegex(
            ValueError,
            'Incorrect allocation',
            allocation.parse_allocations, allocations)
        allocations = ['=,']
        self.assertRaisesRegex(
            ValueError,
            '2 is required',
            allocation.parse_allocations, allocations)
        allocations = ['abc=155']
        self.assertRaisesRegex(
            ValueError,
            'Incorrect allocation',
            allocation.parse_allocations, allocations)
        allocations = ['abc=155,xyz=999']
        self.assertRaisesRegex(
            ValueError,
            'parameter is required',
            allocation.parse_allocations, allocations)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import oslotest.base as base

from osc_placement import version


class TestVersion(base.BaseTestCase):
    def test_compare(self):
        self.assertTrue(version._compare('1.0', version.gt('0.9')))
        self.assertTrue(version._compare('1.0', version.ge('0.9')))
        self.assertTrue(version._compare('1.0', version.ge('1.0')))
        self.assertTrue(version._compare('1.0', version.eq('1.0')))
        self.assertTrue(version._compare('1.0', version.le('1.0')))
        self.assertTrue(version._compare('1.0', version.le('1.1')))
        self.assertTrue(version._compare('1.0', version.lt('1.1')))
        self.assertTrue(
            version._compare('1.1', version.gt('1.0'), version.lt('1.2')))
        self.assertTrue(
            version._compare(
                '0.3', version.eq('0.2'), version.eq('0.3'), op=any))

        # Test error message
        msg = 'Operation or argument is not supported with version 1.0; '
        self.assertEqual((msg + 'requires version greater than 1.0'),
                         version._compare('1.0', version.gt('1.0')))
        self.assertEqual((msg + 'requires at least version 1.1'),
                         version._compare('1.0', version.ge('1.1')))
        self.assertEqual((msg + 'requires version 1.1'),
                         version._compare('1.0', version.eq('1.1')))
        self.assertEqual((msg + 'requires at most version 0.9'),
                         version._compare('1.0', version.le('0.9')))
        self.assertEqual((msg + 'requires version less than 0.9'),
                         version._compare('1.0', version.lt('0.9')))

        self.assertRaises(
            ValueError, version._compare, 'abc', version.le('1.1'))
        self.assertRaises(
            ValueError, version._compare, '1.0', version.le('.0'))
        self.assertRaises(
            ValueError, version._compare, '1', version.le('2'))

        ex = self.assertRaises(
            ValueError, version.compare, '1.0', version.ge('1.1'))
        self.assertEqual(
            'Operation or argument is not supported with version 1.0; '
            'requires at least version 1.1', str(ex))
        ex = self.assertRaises(
            ValueError, version.compare, '1.0',
            version.eq('1.1'), version.eq('1.5'), op=any)
        self.assertEqual(
            'Operation or argument is not supported with version 1.0; '
            'requires version 1.1, or requires version 1.5', str(ex))

    def test_compare_with_exc(self):
        self.assertTrue(version.compare('1.05', version.gt('1.4')))
        self.assertFalse(version.compare('1.3', version.gt('1.4'), exc=False))
        self.assertRaisesRegex(
            ValueError,
            'Operation or argument is not supported',
            version.compare, '3.1.2', version.gt('3.1.3'))

    def test_check_decorator(self):
        fake_api = mock.Mock()
        fake_api_dec = version.check(version.gt('2.11'))(fake_api)
        obj = mock.Mock()
        obj.app.client_manager.placement.api_version = '2.12'
        fake_api_dec(obj, 1, 2, 3)
        fake_api.assert_called_once_with(obj, 1, 2, 3)
        fake_api.reset_mock()
        obj.app.client_manager.placement.api_version = '2.10'
        self.assertRaisesRegex(
            ValueError,
            'Operation or argument is not supported',
            fake_api_dec,
            obj, 1, 2, 3)
        fake_api.assert_not_called()

    def test_check_mixin(self):

        class Test(version.CheckerMixin):
            app = mock.Mock()
            app.client_manager.placement.api_version = '1.2'

        t = Test()
        self.assertTrue(t.compare_version(version.le('1.3')))
        self.assertTrue(t.check_version(version.ge('1.0')))
        self.assertRaisesRegex(
            ValueError,
            'Operation or argument is not supported',
            t.check_version, version.lt('1.2'))

    def test_max_version_consistency(self):
        def _convert_to_tuple(str):
            return tuple(map(int, str.split(".")))

        versions = [
            _convert_to_tuple(ver) for ver in version.SUPPORTED_MICROVERSIONS]
        max_ver = _convert_to_tuple(version.MAX_VERSION_NO_GAP)

        there_is_gap = False
        for i in range(len(versions) - 1):
            j = i + 1
            if versions[j][1] - versions[i][1] != 1:
                there_is_gap = True
                self.assertEqual(max_ver, versions[i])
                break
        if not there_is_gap:
            self.assertEqual(max_ver, versions[-1])

    def test_get_version_returns_max_no_gap_when_no_session(self):
        obj = mock.Mock()
        obj.app.client_manager.session = None
        ret = version.get_version(obj)
        self.assertEqual(version.MAX_VERSION_NO_GAP, ret)
        obj.app.client_manager.placement.api_version.assert_not_called()



# -*- coding: utf-8 -*-
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'openstackdocstheme',
    'sphinx.ext.autodoc',
    'cliff.sphinxext'
]

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'osc-placement'
copyright = '2016, OpenStack Foundation'

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = 'openstackdocs'

# See: https://docs.openstack.org/cliff/2.6.0/sphinxext.html
autoprogram_cliff_application = 'openstack'

autoprogram_cliff_ignored = [
    '--help', '--format', '--column', '--max-width', '--fit-width',
    '--print-empty', '--prefix', '--noindent', '--quote']

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/osc-placement'
openstackdocs_pdf_link = True
openstackdocs_auto_name = False
openstackdocs_use_storyboard = True

# -- Options for LaTeX output -------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_use_xindy = False
latex_documents = [
    ('index', 'doc-osc-placement.tex', 'osc-placement Documentation',
     'OpenStack Foundation', 'manual'),
]
